@@ -1,11 +1,12 @@
+/* 
+ * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* dos emulator, Matthias Lautner
  * Extensions by Robert Sanders, 1992-93
  *
- * $Date: 1995/05/06 16:25:30 $
- * $Source: /usr/src/dosemu0.60/dosemu/RCS/disks.c,v $
- * $Revision: 2.11 $
- * $State: Exp $
- *
  * floppy disks, dos partitions or their images (files) (maximum 8 heads)
  */
 
@@ -29,7 +30,9 @@
 
 #include "emu.h"
 #include "disks.h"
+#include "priv.h"
 
+static int disks_initiated = 0;
 
 #define FDISKS config.fdisks
 #define HDISKS config.hdisks
@@ -97,12 +100,20 @@ int
 read_sectors(struct disk *dp, char *buffer, long head, long sector,
 	     long track, long count)
 {
-  long pos;
+  loff_t  pos;
   long already = 0;
   long tmpread;
 
   /* XXX - header hack. dp->header can be negative for type PARTITION */
   pos = DISK_OFFSET(dp, head, sector, track) + dp->header;
+  d_printf("DISK: Trying to read %ld bytes at T/S/H %ld/%ld/%ld",
+	   count,track,sector,head);
+#ifdef __linux__
+  d_printf(" at pos %Ld\n", pos);
+#else
+  d_printf(" at pos %ld\n", pos);
+#endif
+
 
   /* reads beginning before that actual disk/file */
   if (pos < 0) {
@@ -111,7 +122,7 @@ read_sectors(struct disk *dp, char *buffer, long head, long sector,
     int mbrcount, mbrinc;
 
     if (dp->type != PARTITION) {
-      error("ERROR: negative offset on non-partition disk type\n");
+      error("negative offset on non-partition disk type\n");
       return -DERR_NOTFOUND;
     }
 
@@ -132,14 +143,19 @@ read_sectors(struct disk *dp, char *buffer, long head, long sector,
        * the dest. buffer?  No matter, I guess.
        */
     if ((mbroff + mbrcount) > dp->part_info.mbr_size) {
-      error("ERROR: %s writing in the forbidden fake partition zone!\n",
+      error("%s writing in the forbidden fake partition zone!\n",
 	    dp->dev_name);
       mbrcount = dp->part_info.mbr_size - mbroff;
     }
 
-    d_printf("WARNING: %ld (0x%lx) read for %d below %s, "
+#ifdef __linux__
+    d_printf("WARNING: %Ld",pos);
+#else
+    d_printf("WARNING: %ld",pos);
+#endif
+    d_printf(" (0x%lx) read for %d below %s, "
 	     "h:%ld s:%ld t:%ld\n",
-	     pos, (unsigned long) -pos, readsize,
+	     (unsigned long) -pos, readsize,
 	     dp->dev_name, head, sector, track);
 
     memcpy(buffer, dp->part_info.mbr + mbroff, mbrcount);
@@ -162,8 +178,12 @@ read_sectors(struct disk *dp, char *buffer, long head, long sector,
     }
   }
 
+#ifdef __linux__
+  if (pos != llseek(dp->fdesc, pos, SEEK_SET)) {
+#else
   if (pos != lseek(dp->fdesc, pos, SEEK_SET)) {
-    error("ERROR: Sector not found in read_sector, error = %s!\n",
+#endif
+    error("Sector not found in read_sector, error = %s!\n",
 	  strerror(errno));
     return -DERR_NOTFOUND;
   }
@@ -179,7 +199,7 @@ int
 write_sectors(struct disk *dp, char *buffer, long head, long sector,
 	      long track, long count)
 {
-  off_t pos;
+  loff_t  pos;
   int tmpwrite;
 
   if (dp->rdonly) {
@@ -194,6 +214,13 @@ write_sectors(struct disk *dp, char *buffer, long head, long sector,
 
   /* XXX - header hack */
   pos = DISK_OFFSET(dp, head, sector, track) + dp->header;
+  d_printf("DISK: Trying to write %ld bytes T/S/H %ld/%ld/%ld",
+	   count,track,sector,head);
+#ifdef __linux__
+  d_printf(" at pos %Ld\n", pos);
+#else
+  d_printf(" at pos %ld\n", pos);
+#endif
 
   /* if dp->type is PARTITION, we currently don't allow writing to an area
    * not within the actual partition (i.e. somewhere else on the faked
@@ -204,8 +231,12 @@ write_sectors(struct disk *dp, char *buffer, long head, long sector,
    * Also, I could allow writing to the fake MBR, but I'm not sure that's
    * a good idea.
    */
+#ifdef __linux__
+  if (pos != llseek(dp->fdesc, pos, SEEK_SET)) {
+#else
   if (pos != lseek(dp->fdesc, pos, SEEK_SET)) {
-    error("ERROR: Sector not found in write_sector!\n");
+#endif
+    error("Sector not found in write_sector!\n");
     return -DERR_NOTFOUND;
   }
 
@@ -252,14 +283,16 @@ image_auto(struct disk *dp)
     }
   }
 
-  lseek(dp->fdesc, 0, SEEK_SET);
+  lseek(dp->fdesc, 0, SEEK_SET);/* Use lseek here, as nobody want
+				   > 2^^31 bytes for the image file */
   if (RPT_SYSCALL(read(dp->fdesc, header, HEADER_SIZE)) != HEADER_SIZE) {
-    error("ERROR: could not read full header in image_init\n");
+    error("could not read full header in image_init\n");
     leavedos(19);
   }
 
-  if (strncmp(header, IMAGE_MAGIC, IMAGE_MAGIC_SIZE)) {
-    error("ERROR: IMAGE %s header lacks magic string - cannot autosense!\n",
+  if (strncmp(header, IMAGE_MAGIC, IMAGE_MAGIC_SIZE)
+		&& (*((long *)header) != DEXE_MAGIC) ) {
+    error("IMAGE %s header lacks magic string - cannot autosense!\n",
 	  dp->dev_name);
     leavedos(20);
   }
@@ -283,7 +316,7 @@ hdisk_auto(struct disk *dp)
 
   if (ioctl(dp->fdesc, DIOCGDINFO, &label) != 0 ||
       fstat(dp->fdesc, &stb) != 0) {
-      error("ERROR: can't fstat %s: %s\n", dp->dev_name,
+      error("can't fstat %s: %s\n", dp->dev_name,
 	    strerror(errno));
       leavedos(21);
   } else {
@@ -299,7 +332,7 @@ hdisk_auto(struct disk *dp)
   struct hd_geometry geo;
 
   if (ioctl(dp->fdesc, HDIO_GETGEO, &geo) < 0) {
-    error("ERROR: can't get GEO of %s: %s\n", dp->dev_name,
+    error("can't get GEO of %s: %s\n", dp->dev_name,
 	  strerror(errno));
     leavedos(21);
   }
@@ -326,6 +359,7 @@ hdisk_auto(struct disk *dp)
 void
 partition_setup(struct disk *dp)
 {
+  PRIV_SAVE_AREA
   int part_fd, i;
   unsigned char tmp_mbr[SECTOR_SIZE];
   char *hd_name;
@@ -351,8 +385,11 @@ partition_setup(struct disk *dp)
   hd_name[8] = '\0';			/* i.e.  /dev/hda6 -> /dev/hda */
 #endif
 
-  if ((part_fd = DOS_SYSCALL(open(hd_name, O_RDONLY))) == -1) {
-    error("ERROR: opening device %s to read MBR for PARTITION %s\n",
+  enter_priv_on();
+  part_fd = DOS_SYSCALL(open(hd_name, O_RDONLY));
+  leave_priv_setting();
+  if (part_fd == -1) {
+    error("opening device %s to read MBR for PARTITION %s\n",
 	  hd_name, dp->dev_name);
     leavedos(22);
   }
@@ -390,7 +427,7 @@ partition_setup(struct disk *dp)
 			     dp->part_info.beg_cyl) +
 		 (SECTOR_SIZE * (dp->part_info.pre_secs - 1)));
 #else
-  dp->header = -(SECTOR_SIZE * (dp->part_info.pre_secs));
+  dp->header = -(SECTOR_SIZE * (loff_t) (dp->part_info.pre_secs));
 #endif
 
   dp->part_info.mbr_size = SECTOR_SIZE;
@@ -442,7 +479,7 @@ set_part_ent(struct disk *dp, char *tmp_mbr)
 
   if (ioctl(dp->fdesc, DIOCGDINFO, &label) != 0 ||
       fstat(dp->fdesc, &stb) != 0) {
-      error("ERROR: can't fstat %s: %s\n", dp->dev_name,
+      error("can't fstat %s: %s\n", dp->dev_name,
 	    strerror(errno));
       leavedos(21);
   } else
@@ -450,7 +487,7 @@ set_part_ent(struct disk *dp, char *tmp_mbr)
 #endif
 #ifdef __linux__
   if (ioctl(dp->fdesc, BLKGETSIZE, &length)) {
-    error("ERROR: calling ioctl BLKGETSIZE for PARTITION %s\n", dp->dev_name);
+    error("calling ioctl BLKGETSIZE for PARTITION %s\n", dp->dev_name);
     leavedos(22);
   }
 #endif
@@ -493,6 +530,7 @@ disk_close(void)
 {
   struct disk *dp;
 
+  if (!disks_initiated) return;  /* just to be safe */
   for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
     if (dp->removeable && dp->fdesc >= 0) {
       d_printf("DISK: Closing a disk\n");
@@ -506,15 +544,20 @@ disk_close(void)
 void
 disk_open(struct disk *dp)
 {
+  PRIV_SAVE_AREA
   struct fd_type fl;
 
   if (dp == NULL || dp->fdesc >= 0)
     return;
     
+  enter_priv_on();
   dp->fdesc = DOS_SYSCALL(open(dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0));
+  leave_priv_setting();
   if (dp->fdesc < 0) 
     if (errno == EROFS || errno == EACCES) {
+      enter_priv_on();
       dp->fdesc = DOS_SYSCALL(open(dp->dev_name, O_RDONLY, 0));
+      leave_priv_setting();
       if (dp->fdesc < 0) {
         d_printf("ERROR: (disk) can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
         /* In case we DO get more clever, we want to share that code */
@@ -543,7 +586,7 @@ disk_open(struct disk *dp)
       dp->tracks = 0;
       return;
     }
-    error("ERROR: can't get floppy parameter of %s (%s)\n", dp->dev_name, strerror(err));
+    error("can't get floppy parameter of %s (%s)\n", dp->dev_name, strerror(err));
     fatalerr = 5;
     return;
   }
@@ -573,15 +616,21 @@ static void sigalarm_onoff(int on)
 void
 disk_open(struct disk *dp)
 {
+  PRIV_SAVE_AREA
   struct floppy_struct fl;
 
   if (dp == NULL || dp->fdesc >= 0)
     return;
     
-  dp->fdesc = DOS_SYSCALL(open(dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0));
-  if (dp->fdesc < 0) 
-    if (errno == EROFS) {
+  enter_priv_on();
+  dp->fdesc = SILENT_DOS_SYSCALL(open(dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0));
+  leave_priv_setting();
+
+  if (dp->fdesc < 0) {
+    if (errno == EROFS || errno == ENODEV) {
+      enter_priv_on();
       dp->fdesc = DOS_SYSCALL(open(dp->dev_name, O_RDONLY, 0));
+      leave_priv_setting();
       if (dp->fdesc < 0) {
         d_printf("ERROR: (disk) can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
         /* In case we DO get more clever, we want to share that code */
@@ -599,6 +648,7 @@ disk_open(struct disk *dp)
 #endif
       return;
     }
+  }
   else dp->rdonly = dp->wantrdonly;
 
 {
@@ -624,7 +674,7 @@ disk_open(struct disk *dp)
       dp->tracks = 0;
       return;
     }
-    error("ERROR: can't get floppy parameter of %s (%s)\n", dp->dev_name, sys_errlist[errno]);
+    error("can't get floppy parameter of %s (%s)\n", dp->dev_name, sys_errlist[errno]);
     fatalerr = 5;
     return;
   }
@@ -641,10 +691,12 @@ void
 disk_close_all(void)
 {
   struct disk *dp;
-
+  
+  if (!disks_initiated) return;  /* prevent idiocy */
   if (config.bootdisk && bootdisk.fdesc >= 0) {
+    d_printf("Boot disk Closing %x\n", bootdisk.fdesc);
     (void) close(bootdisk.fdesc);
-    bootdisk.fdesc = 0;
+    bootdisk.fdesc = -1;
     d_printf("BOOTDISK Closing\n");
   }
   for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
@@ -661,6 +713,7 @@ disk_close_all(void)
       dp->fdesc = -1;
     }
   }
+  disks_initiated = 0;
 }
 
 /*
@@ -674,6 +727,7 @@ disk_close_all(void)
 void
 disk_init(void)
 {
+  PRIV_SAVE_AREA
   struct disk *dp=NULL;
   struct stat stbuf;
 
@@ -683,29 +737,35 @@ disk_init(void)
 
 #endif
 
+  disks_initiated = 1;  /* disk_init has been called */
   if (config.bootdisk) {
+    enter_priv_on();
     bootdisk.fdesc = open(bootdisk.dev_name,
 			  bootdisk.rdonly ? O_RDONLY : O_RDWR, 0);
-    if (bootdisk.fdesc < 0) 
+    leave_priv_setting();
+    if (bootdisk.fdesc < 0) {
       if (errno == EROFS) {
+        enter_priv_on();
         bootdisk.fdesc = open(bootdisk.dev_name, O_RDONLY, 0);
+        leave_priv_setting();
         if (bootdisk.fdesc < 0) {
-          error("ERROR: can't open bootdisk %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
+          error("can't open bootdisk %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
           leavedos(23);
         } else {
           bootdisk.rdonly = 1;
           d_printf("(disk) can't open bootdisk %s for read/write. Readonly did work though\n", bootdisk.dev_name);
         }
       } else {
-        error("ERROR: can't open bootdisk %s: %sn", dp->dev_name, strerror(errno));
+        error("can't open bootdisk %s: %sn", dp->dev_name, strerror(errno));
         leavedos(23);
       }
+    }
     else bootdisk.rdonly = bootdisk.wantrdonly;
     bootdisk.removeable = 0;
   }
   for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
     if (stat(dp->dev_name, &stbuf) < 0) {
-      error("ERROR: can't stat %s\n", dp->dev_name);
+      error("can't stat %s\n", dp->dev_name);
       leavedos(24);
     }
     if (S_ISBLK(stbuf.st_mode))
@@ -730,50 +790,62 @@ disk_init(void)
       continue;
     }
 #endif
+    enter_priv_on();
     dp->fdesc = open(dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
-    if (dp->fdesc < 0) 
+    leave_priv_setting();
+    if (dp->fdesc < 0) {
       if (errno == EROFS || errno == EACCES) {
+        enter_priv_on();
         dp->fdesc = open(dp->dev_name, O_RDONLY, 0);
+        leave_priv_setting();
         if (dp->fdesc < 0) {
-          error("ERROR: can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
+          error("can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
           leavedos(25);
         } else {
           dp->rdonly = 1;
           d_printf("(disk) can't open %s for read/write. Readonly did work though\n", dp->dev_name);
         }
       } else {
-        error("ERROR: can't open %s: %s\n", dp->dev_name, strerror(errno));
+        error("can't open %s: %s\n", dp->dev_name, strerror(errno));
         leavedos(25);
       }
+    }
     else dp->rdonly = dp->wantrdonly;
   }
   for (dp = hdisktab; dp < &hdisktab[HDISKS]; dp++) {
     if(dp->type == IMAGE)  {
-	priv_off();
-	d_printf("IMAGE: Using user permissions\n");
+	if (dp->dexeflags & DISK_DEXE_RDWR) {
+	  d_printf("IMAGE: dexe, RDWR access allowed for %s\n",dp->dev_name);
+	  enter_priv_on();
+	}
+	else {
+	  d_printf("IMAGE: Using user permissions\n");
+	  enter_priv_off();
+	}
     }
+    else enter_priv_on();
     dp->fdesc = open(dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
+    leave_priv_setting();
     if (dp->fdesc < 0) {
       if (errno == EROFS || errno == EACCES) {
+        if (dp->type == IMAGE) enter_priv_off(); else enter_priv_on();
         dp->fdesc = open(dp->dev_name, O_RDONLY, 0);
+        leave_priv_setting();
         if (dp->fdesc < 0) {
-          error("ERROR: can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
+          error("can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
           leavedos(26);
         } else {
           dp->rdonly = 1;
           d_printf("(disk) can't open %s for read/write. Readonly did work though\n", dp->dev_name);
         }
       } else {
-        error("ERROR: can't open %s: #%d - %s\n", dp->dev_name, errno, strerror(errno));
+        error("can't open %s: #%d - %s\n", dp->dev_name, errno, strerror(errno));
         leavedos(26);
       }
     }
     else dp->rdonly = dp->wantrdonly;
     dp->removeable = 0;
 
-    if(dp->type == IMAGE)
-	priv_on();
-
     /* HACK: if unspecified geometry (-1) then try to get it from kernel.
        May only work on WD compatible disks (MFM/RLL/ESDI/IDE). */
     if (dp->sectors == -1)
@@ -793,8 +865,7 @@ disk_init(void)
      */
 #ifdef SILLY_GET_GEOMETRY
     if (RPT_SYSCALL(read(dp->fdesc, buf, 512)) != 512) {
-      error("ERROR: can't read disk info of %s\n", dp->dev_name);
-      priv_on();
+      error("can't read disk info of %s\n", dp->dev_name);
       leavedos(27);
     }
 
@@ -825,8 +896,7 @@ disk_init(void)
 	     *(unsigned long *) &buf[39], label);
 
     if (s % (dp->sectors * dp->heads) != 0) {
-      error("ERROR: incorrect track number of %s\n", dp->dev_name);
-      priv_on();
+      error("incorrect track number of %s\n", dp->dev_name);
       /* leavedos(28); */
     }
 #endif
@@ -853,9 +923,10 @@ int13(u_char i)
 {
   unsigned int disk, head, sect, track, number;
   int res;
-  long pos;
+  loff_t  pos;
   char *buffer;
   struct disk *dp;
+  int checkdp_val;
 
   disk = LO(dx);
   if (disk < FDISKS) {
@@ -863,6 +934,15 @@ int13(u_char i)
       dp = &bootdisk;
     else
       dp = &disktab[disk];
+    switch (HI(ax)) {
+      #define DISKETTE_MOTOR_TIMEOUT (*((unsigned char *)0x440))
+      /* NOTE: we don't need this counter, but older games seem to rely
+       * on it. We count it down in INT08 (bios.S) --SW, --Hans
+       */
+      case 0: case 2: case 3: case 5: case 10: case 11:
+        DISKETTE_MOTOR_TIMEOUT = 3*18;  /* set timout to 3 seconds */
+        break;
+    }
   }
   else if (disk >= 0x80 && disk < 0x80 + HDISKS)
     dp = &hdisktab[disk - 0x80];
@@ -889,10 +969,16 @@ int13(u_char i)
   case 2:			/* read */
     FLUSHDISK(dp);
     disk_open(dp);
-    head = HI(dx);
+    checkdp_val = checkdp(dp);
+    if (!checkdp_val && dp->diskcyl4096 && dp->heads <= 64)
+      head = HI(dx) & 0x3f;
+    else
+      head = HI(dx);
     sect = (REG(ecx) & 0x3f) - 1;
     track = (HI(cx)) |
       ((REG(ecx) & 0xc0) << 2);
+    if (!checkdp_val && dp->diskcyl4096 && dp->heads <= 64 && (HI(dx) & 0xc0))
+      track |= (HI(dx) & 0xc0) << 4;
     buffer = SEG_ADR((char *), es, bx);
     number = LO(ax);
 #if 0
@@ -900,9 +986,9 @@ int13(u_char i)
 	     disk, head, sect, track, number, (void *) buffer);
 #endif
 
-    if (checkdp(dp) || head >= dp->heads ||
+    if (checkdp_val || head >= dp->heads ||
 	sect >= dp->sectors || track >= dp->tracks) {
-      error("ERROR: Sector not found 1!\n");
+      error("Sector not found 1!\n");
       d_printf("DISK %d read [h:%d,s:%d,t:%d](%d)->%p\n",
 	       disk, head, sect, track, number, (void *) buffer);
       if (dp) {
@@ -926,7 +1012,7 @@ int13(u_char i)
       break;
     }
     else if (res & 511) {	/* must read multiple of 512 bytes */
-      error("ERROR: sector_corrupt 1, return = %d!\n", res);
+      error("sector_corrupt 1, return = %d!\n", res);
       HI(ax) = DERR_BADSEC;	/* sector corrrupt */
       CARRY;
       break;
@@ -941,18 +1027,24 @@ int13(u_char i)
   case 3:			/* write */
     FLUSHDISK(dp);
     disk_open(dp);
-    head = HI(dx);
+    checkdp_val = checkdp(dp);
+    if (!checkdp_val && dp->diskcyl4096 && dp->heads <= 64)
+      head = HI(dx) & 0x3f;
+    else
+      head = HI(dx);
     sect = (REG(ecx) & 0x3f) - 1;
     track = (HI(cx)) |
       ((REG(ecx) & 0xc0) << 2);
+    if (!checkdp_val && dp->diskcyl4096 && dp->heads <= 64 && (HI(dx) & 0xc0))
+      track |= (HI(dx) & 0xc0) << 4;
     buffer = SEG_ADR((char *), es, bx);
     number = LO(ax);
     W_printf("DISK write [h:%d,s:%d,t:%d](%d)->%p\n",
 	     head, sect, track, number, (void *) buffer);
 
-    if (checkdp(dp) || head >= dp->heads ||
+    if (checkdp_val || head >= dp->heads ||
 	sect >= dp->sectors || track >= dp->tracks) {
-      error("ERROR: Sector not found 3!\n");
+      error("Sector not found 3!\n");
       show_regs(__FILE__, __LINE__);
       HI(ax) = DERR_NOTFOUND;
       REG(eflags) |= CF;
@@ -960,7 +1052,7 @@ int13(u_char i)
     }
 
     if (dp->rdonly) {
-      error("ERROR: write protect!\n");
+      error("write protect!\n");
       show_regs(__FILE__, __LINE__);
       if (dp->removeable)
 	HI(ax) = DERR_WP;
@@ -981,7 +1073,7 @@ int13(u_char i)
       break;
     }
     else if (res & 511) {	/* must write multiple of 512 bytes */
-      error("ERROR: Write sector corrupt 2 (wrong size)!\n");
+      error("Write sector corrupt 2 (wrong size)!\n");
       HI(ax) = DERR_BADSEC;
       CARRY;
       break;
@@ -1008,19 +1100,27 @@ int13(u_char i)
 	sect >= dp->sectors || track >= dp->tracks) {
       HI(ax) = DERR_NOTFOUND;
       REG(eflags) |= CF;
-      error("ERROR: test: sector not found 5\n");
+      error("test: sector not found 5\n");
       dbug_printf("hds: %d, sec: %d, tks: %d\n",
 		  dp->heads, dp->sectors, dp->tracks);
       break;
     }
+#ifdef __linux__
+    pos = (long long ) ((track * dp->heads + head) * dp->sectors + sect) << 9;
+#else
     pos = (long) ((track * dp->heads + head) * dp->sectors + sect) << 9;
+#endif
     /* XXX - header hack */
     pos += dp->header;
 
+#ifdef __linux__
+    if (pos != llseek(dp->fdesc, pos, 0)) {
+#else
     if (pos != lseek(dp->fdesc, pos, 0)) {
+#endif
       HI(ax) = DERR_NOTFOUND;
       REG(eflags) |= CF;
-      error("ERROR: test: sector not found 6\n");
+      error("test: sector not found 6\n");
       break;
     }
 #if 0
@@ -1028,7 +1128,7 @@ int13(u_char i)
     if (res & 0x1ff) {		/* must read multiple of 512 bytes  and res != -1 */
       HI(ax) = DERR_BADSEC;
       REG(eflags) |= CF;
-      error("ERROR: test: sector corrupt 3\n");
+      error("test: sector corrupt 3\n");
       break;
     }
     LWORD(eax) = res >> 9;
@@ -1073,7 +1173,7 @@ int13(u_char i)
       HI(cx) = (dp->tracks - 1) & 0xff;
 
       LO(dx) = (disk < 0x80) ? FDISKS : HDISKS;
-      LO(cx) = dp->sectors | ((dp->tracks & 0x300) >> 2);
+      LO(cx) = dp->sectors | (((dp->tracks -1) & 0x300) >> 2);
       LO(ax) = 0;
       HI(ax) = DERR_NOERR;
       REG(eflags) &= ~CF;	/* no error */
@@ -1167,7 +1267,7 @@ int13(u_char i)
 	NOCARRY;
       }
       else {
-	error("ERROR: gettype: no disk %d\n", disk);
+	error("gettype: no disk %d\n", disk);
 	HI(ax) = 0;		/* disk not there */
 	REG(eflags) |= CF;	/* error */
       }
@@ -1214,6 +1314,180 @@ int13(u_char i)
   case 0x28:			/* DRDOS 6.0 call ??? */
     d_printf("int 13h, ax=%04x...DRDOS call\n", LWORD(eax));
     break;
+
+  case 0x41:                    /* IBM/MS Extensions, install check */
+    LWORD(ebx) = IMEXT_MAGIC;
+    HI(ax) = IMEXT_VER_MAJOR;
+    LWORD(ecx) = IMEXT_API_SUPPORT_BITS;
+    NOCARRY;
+    break;
+
+  case 0x42: {                  /* IBM/MS Extensions, read */
+    struct ibm_ms_diskaddr_pkt *diskaddr;
+
+    FLUSHDISK(dp);
+    disk_open(dp);
+    diskaddr = SEG_ADR((struct ibm_ms_diskaddr_pkt *), ds, si);
+    sect = diskaddr->block_lo % dp->sectors;
+    head = (diskaddr->block_lo / dp->sectors) % dp->heads;
+    track = diskaddr->block_lo / (dp->heads * dp->sectors);
+    buffer = (char *)(((unsigned long)diskaddr->buf_seg << 4)+diskaddr->buf_ofs);
+    number = diskaddr->blocks;
+    diskaddr->blocks = 0;
+
+    if (checkdp(dp) || track >= dp->tracks) {
+      error("ERROR: Sector not found, AH=0x42!\n");
+      d_printf("DISK %d ext read [h:%d,s:%d,t:%d](%d)->%p\n",
+	       disk, head, sect, track, number, (void *) buffer);
+      if (dp) {
+	  d_printf("DISK dev %s GEOM %d heads %d sects %d trk\n",
+		   dp->dev_name, dp->heads, dp->sectors, dp->tracks);
+      } else {
+	  d_printf("DISK %x undefined.\n", disk);
+      }
+
+      HI(ax) = DERR_NOTFOUND;
+      REG(eflags) |= CF;
+      show_regs(__FILE__, __LINE__);
+      break;
+    }
+
+    res = read_sectors(dp, buffer, head, sect, track, number);
+
+    if (res < 0) {
+      HI(ax) = -res;
+      CARRY;
+      break;
+    }
+    else if (res & 511) {	/* must read multiple of 512 bytes */
+      error("ERROR: sector_corrupt 1, return = %d!\n", res);
+      HI(ax) = DERR_BADSEC;	/* sector corrrupt */
+      CARRY;
+      break;
+    }
+
+    diskaddr->blocks = res >> 9;
+    HI(ax) = 0;
+    REG(eflags) &= ~CF;
+    R_printf("DISK ext read @%d/%d/%d (%d) -> %p OK.\n",
+	     head, track, sect, res >> 9, (void *)buffer);
+    break;
+  }
+  
+  case 0x43: {                  /* IBM/MS Extensions, write */
+    struct ibm_ms_diskaddr_pkt *diskaddr;
+
+    FLUSHDISK(dp);
+    disk_open(dp);
+    diskaddr = SEG_ADR((struct ibm_ms_diskaddr_pkt *), ds, si);
+    sect = diskaddr->block_lo % dp->sectors;
+    head = (diskaddr->block_lo / dp->sectors) % dp->heads;
+    track = diskaddr->block_lo / (dp->heads * dp->sectors);
+    buffer = (char *)(((unsigned long)diskaddr->buf_seg << 4)+diskaddr->buf_ofs);
+    number = diskaddr->blocks;
+    diskaddr->blocks = 0;
+
+    if (checkdp(dp) || track >= dp->tracks) {
+      error("ERROR: Sector not found, AH=0x42!\n");
+      d_printf("DISK %d ext write [h:%d,s:%d,t:%d](%d)->%p\n",
+	       disk, head, sect, track, number, (void *) buffer);
+      if (dp) {
+	  d_printf("DISK dev %s GEOM %d heads %d sects %d trk\n",
+		   dp->dev_name, dp->heads, dp->sectors, dp->tracks);
+      } else {
+	  d_printf("DISK %x undefined.\n", disk);
+      }
+
+      HI(ax) = DERR_NOTFOUND;
+      REG(eflags) |= CF;
+      show_regs(__FILE__, __LINE__);
+      break;
+    }
+
+    if (dp->rdonly) {
+      d_printf("DISK is write protected!\n");
+      show_regs(__FILE__, __LINE__);
+      if (dp->removeable)
+	HI(ax) = DERR_WP;
+      else
+	HI(ax) = DERR_WRITEFLT;
+      REG(eflags) |= CF;
+      break;
+    }
+
+    if (dp->rdonly)
+      error("CONTINUED!!!!!\n");
+    res = write_sectors(dp, buffer, head, sect, track, number);
+
+    if (res < 0) {
+      HI(ax) = -res;
+      CARRY;
+      break;
+    }
+    else if (res & 511) {	/* must read multiple of 512 bytes */
+      error("ERROR: sector_corrupt 1, return = %d!\n", res);
+      HI(ax) = DERR_BADSEC;	/* sector corrrupt */
+      CARRY;
+      break;
+    }
+
+    diskaddr->blocks = res >> 9;
+    HI(ax) = 0;
+    REG(eflags) &= ~CF;
+    R_printf("DISK ext write @%d/%d/%d (%d) -> %p OK.\n",
+	     head, track, sect, res >> 9, (void *)buffer);
+    break;
+  }
+
+  case 0x44:                    /* IBM/MS Extensions, verify */
+    /* Always succeeds. Should perhaps check validity of sector address. */
+    NOCARRY;
+    HI(ax) = 0;
+    break;
+
+  case 0x47:                    /* IBM/MS Extensions, extended seek */
+    NOCARRY;
+    HI(ax) = 0;
+    break;
+
+  case 0x48: {                  /* IBM/MS Extensions, get drive parameters */
+    struct ibm_ms_drive_params *params;
+
+    FLUSHDISK(dp);
+    disk_open(dp);
+    params = SEG_ADR((struct ibm_ms_drive_params *), ds, si);
+
+    if (checkdp(dp)) {
+      error("ERROR: Invalid drive, AH=0x48!\n");
+      if (dp) {
+	  d_printf("DISK dev %s GEOM %d heads %d sects %d trk\n",
+		   dp->dev_name, dp->heads, dp->sectors, dp->tracks);
+      } else {
+	  d_printf("DISK %x undefined.\n", disk);
+      }
+
+      HI(ax) = DERR_NOTFOUND;
+      REG(eflags) |= CF;
+      show_regs(__FILE__, __LINE__);
+      break;
+    }
+
+    params->flags = IMEXT_INFOFLAG_CHSVALID;
+    if (dp->removeable)
+      params->flags |= IMEXT_INFOFLAG_REMOVABLE;
+    params->tracks = dp->tracks;
+    params->heads = dp->heads;
+    params->sectors = dp->sectors;
+    params->total_sectors_lo = dp->tracks*dp->heads*dp->sectors;
+    params->total_sectors_hi = 0;
+    params->bytes_per_sector = SECTOR_SIZE;
+    if (params->len >= 0x1e)
+      params->edd_cfg_ofs = params->edd_cfg_seg = 0xffff;
+    NOCARRY;
+    HI(ax) = 0;
+    break;
+  }
+
   case 0x5:			/* format */
     NOCARRY;			/* successful */
     HI(ax) = DERR_NOERR;
@@ -1230,7 +1504,7 @@ int13(u_char i)
     NOCARRY;
     break;
   default:
-    error("ERROR: disk error, unknown command: int13, ax=0x%x\n",
+    error("disk error, unknown command: int13, ax=0x%x\n",
 	  LWORD(eax));
     show_regs(__FILE__, __LINE__);
     CARRY;
@@ -1253,3 +1527,4 @@ floppy_tick(void)
     secs = 0;
   }
 }
+
