@@ -1,13 +1,12 @@
+/* 
+ * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* mouse.c for the DOS emulator
  *       Robert Sanders, gt8134b@prism.gatech.edu
  *
- * $Date: 1995/05/06 16:27:42 $
- * $Source: /usr/src/dosemu0.60/mouse/RCS/mouse.c,v $
- * $Revision: 2.20 $
- * $State: Exp $
- *
- * $Log: mouse.c,v $
- *
  */
 
 #include <stdio.h>
@@ -36,6 +35,11 @@
 
 #include "vc.h"
 #include "port.h"
+
+#include "keyboard.h"
+
+#include "priv.h"
+
 extern int get_perm ();
 extern int release_perm ();
 
@@ -51,15 +55,31 @@ extern void X_change_mouse_cursor(void);
 void DOSEMUSetupMouse(void);
 
 #define Mouse_INT	(0x33 * 16)
+#ifndef USE_NEW_INT
 /* I have not properly tested this INT74 - JES 96/10/20 */
 #define Mouse_INT74	(0x74 * 16)
+#else /* USE_NEW_INT */
+/* DANG_BEGIN_REMARK
+ * I have not properly tested this INT74 - JES 96/10/20 
+ * I have removed it.  INT74 is irq 12.  Which I suppose is the proper
+ * irq for a ps2 mouse.  It appears initial support was planned to
+ * support irq 12 and at Mouse_ROUTINE_OFF is a routine that
+ * acknowledges an irq.  That routine is probably what should be
+ * acknowledging irq12, and what int 0x74 should point to.  
+ * I have disabled int0x74 support for now. --EB 29 Oct 1997
+ * DANG_END_REMARK
+ */
+#endif /* USE_NEW_INT */
 
 static
 void mouse_cursor(int), mouse_pos(void), mouse_setpos(void),
  mouse_setxminmax(void), mouse_setyminmax(void), mouse_set_tcur(void),
  mouse_set_gcur(void), mouse_setsub(void), mouse_bpressinfo(void), mouse_brelinfo(void),
  mouse_mickeys(void), mouse_version(void), mouse_enable_internaldriver(void),
- mouse_disable_internaldriver(void), mouse_software_reset(void), 
+ mouse_disable_internaldriver(void),
+#if 0
+ mouse_software_reset(void), 
+#endif
  mouse_getgeninfo(void), mouse_exclusionarea(void), mouse_setcurspeed(void),
  mouse_undoc1(void), mouse_storestate(void), mouse_restorestate(void),
  mouse_getmaxcoord(void), mouse_getstorereq(void), mouse_setsensitivity(void),
@@ -152,7 +172,13 @@ mouse_helper(void)
   case 0:				/* Reset iret for mouse */
     m_printf("MOUSE move iret !\n");
     SETIVEC(0x33, Mouse_SEG, Mouse_INT);
+#ifndef USE_NEW_INT
     SETIVEC(0x74, Mouse_SEG, Mouse_INT74);
+#else /* USE_NEW_INT */
+#if 0
+    SETIVEC(0x74, Mouse_SEG, Mouse_ROUTINE_OFF);
+#endif
+#endif /* USE_NEW_INT */
     break;
   case 1:				/* Select Microsoft Mode */
     m_printf("MOUSE Microsoft Mouse (two buttons) selected.\n");
@@ -561,6 +587,7 @@ mouse_getgeninfo(void)
   LWORD(eax) |= 0x8000;			/* Set SYS file */
 }
 
+#if 0
 static void
 mouse_software_reset(void)
 {
@@ -571,7 +598,13 @@ mouse_software_reset(void)
   mouse.cs=0;
   mouse.ip=0;
   SETIVEC(0x33, Mouse_SEG, Mouse_INT);
+#ifndef USE_NEW_INT
   SETIVEC(0x74, Mouse_SEG, Mouse_INT74);
+#else /* USE_NEW_INT */
+#if 0
+  SETIVEC(0x74, Mouse_SEG, Mouse_ROUTINE_OFF);
+#endif
+#endif /* USE_NEW_INT */
 
 #ifdef X_SUPPORT
 #ifndef X_SLOW_CHANGE_CURSOR
@@ -589,6 +622,7 @@ mouse_software_reset(void)
   else
     LWORD(ebx) = 2;
 }
+#endif
 
 static void
 mouse_detsensitivity(void)
@@ -621,11 +655,11 @@ mouse_restorestate(void)
   	mouse_cursor(-1);
   }
 
-  memcpy(&mouse, (u_char *)(LWORD(es) << 4)+LWORD(edx), sizeof(mouse));
+  memcpy((void *)&mouse, (u_char *)(LWORD(es) << 4)+LWORD(edx), sizeof(mouse));
 
   /* regenerate mouse graphics cursor from masks; they take less
   	space than the "compiled" version. */
-  define_graphics_cursor(mouse.graphscreenmask,mouse.graphcursormask);
+  define_graphics_cursor((short *)mouse.graphscreenmask,(short *)mouse.graphcursormask);
 
   /* we turned off the mouse cursor prior to saving, so turn it
   	back on again at the restore. */
@@ -650,7 +684,7 @@ mouse_storestate(void)
   	mouse_cursor(-1);
   }
 
-  memcpy((u_char *)(LWORD(es) << 4)+LWORD(edx), &mouse, sizeof(mouse));
+  memcpy((u_char *)(LWORD(es) << 4)+LWORD(edx), (void *)&mouse, sizeof(mouse));
 
   /* now turn it back on */
   if (mouse.cursor_on >= 0) {
@@ -710,8 +744,13 @@ mouse_setcurspeed(void)
   }
 }
 
-
-static void
+/*
+ * Because the mouse hook finally works all video sets that pass
+ * through the video bios eventually pass through here.
+ * Win31 does not reset the mouse in the normal way, and
+ * the only clue we have is the video mode (0x62 and 0x06 for 1024x768)
+ */
+void
 mouse_reset_to_current_video_mode(void)
 {
   /* Setup MAX / MIN co-ordinates */
@@ -723,40 +762,93 @@ mouse_reset_to_current_video_mode(void)
   */
   get_current_video_mode();
 
+  /* 
+   * Actually what happens is: if a Text mode is found, height and width
+   * are in characters, else they are in pixels. This was clearly done to
+   * confuse people. Besides, they ignore the ACTUAL maxx,maxy values as
+   * stored in the BIOS variables.
+   */
+  /* To get quicken to run correctly I no longer ignore maxx and maxy
+   * values as stored in the BIOS variables for text mode.
+   * Does this violate a spec?
+   * -- EB 25 May 1998
+   */
+
+/* DANG_BEGIN_REMARK
+ * Whoever wrote the dos mouse driver spec was brain dead...
+ * For some video modes the mouse driver appears to randomly
+ * pick a shift factor, possibly to keep at least a 640x200 resolution.
+ *
+ * The general programming documentation doesn't make this clear.
+ * And says that in text modes it is safe to divide the resolution by 
+ * 8 to get the coordinates in characters.
+ *
+ * The only safe way to handle the mouse driver is to call function
+ * 0x26 Get max x & max y coordinates and scale whatever the driver 
+ * returns yourself.
+ *
+ * To handle programs written by programmers who weren't so cautious a
+ * doctrine of least suprise has been implemented.
+ *
+ * As much as possible do the same as a standard dos mouse driver in the
+ * original vga modes 0,1,2,3,4,5,6,7,13,14,15,16,17,18,19. 
+ *
+ * For other text modes allow the divide by 8 technique to work.
+ * For other graphics modes return x & y in screen coordinates.
+ * Except when those modes are either 40x?? or 320x???
+ * and then handle the x resolution as in 40x25 and 320x200 modes.
+ *
+ * 320x200 modes are slightly controversial as I have indications that
+ * not all mouse drivers do the same thing. So I have taken the
+ * simplest, and most common route, which is also long standing dosemu
+ * practice of always shifting by 1.
+ *
+ * -- Eric Biederman 19 August 1998
+ * DANG_END_REMARK
+ */
+
   if (current_video.textgraph == 'T') {
+    mouse.xshift = 3;
+    mouse.yshift = 3;
+    if (current_video.width == 40) {
+	    mouse.xshift = 4;
+    }
     mouse.gfx_cursor = FALSE;
   } else {
     mouse.gfx_cursor = TRUE;
+    mouse.xshift = 0;    
+    mouse.yshift = 0;
+    if (current_video.width == 320) {
+	    mouse.yshift = 1;
+	    /* Every working example I can find picks a yshift of 1 in
+	     * 320x200 mode.  Unless some real prominent mouse driver
+	     * handles this mode differently this should be fine.
+	     * -- Eric Biederman 19 August 1998
+	     */
+    }
     define_graphics_cursor(default_graphscreenmask,default_graphcursormask);
   }
 
-  /* virtual resolution is always at least 640x200 */
-  mouse.maxx = 640;
-  mouse.maxy = current_video.height;
-  if (mouse.maxy < 200) {
-  	mouse.maxy = 200;
-  	mouse.yshift = 3;
-  }
-  else
-  	mouse.yshift = 0;
+  /* Set the maximum sizes */
+  mouse.maxx = current_video.width << mouse.xshift;
+  mouse.maxy = current_video.height << mouse.yshift;
 
+  /* Find the center of the screen */
   mouse.x = (mouse.minx + mouse.maxx) / 2;
   mouse.y = (mouse.miny + mouse.maxy) / 2;
 
   /* force update first time after reset */
   mouse.oldrx = -1;
-
-  /* determine shift compensation factors between physical and
-  	virtual resolution */
-  switch (current_video.width) {
-  	case 40: mouse.xshift = 4; break;
-  	case 80: mouse.xshift = 3; break;
-  	case 320: mouse.xshift = 1; break;
-  	default: mouse.xshift = 0; break;
-  }
-
+  
+  /* Change mouse.maxx & mouse.maxy from ranges to maximums */
   mouse.maxx = mouse_roundx(mouse.maxx - 1);
   mouse.maxy = mouse_roundy(mouse.maxy - 1);
+  mouse.maxx += (1 << mouse.xshift) -1;
+  mouse.maxy += (1 << mouse.yshift) -1;
+
+  m_printf("maxx=%i, maxy=%i speed_x=%i speed_y=%i ignorexy=%i type=%d\n", 
+	   mouse.maxx, mouse.maxy, mouse.speed_x, mouse.speed_y, mouse.ignorexy,
+	   mice->type);
 }
 
 
@@ -765,7 +857,13 @@ mouse_reset(int flag)
 {
   m_printf("MOUSE: reset mouse/installed!\n");
   SETIVEC(0x33, Mouse_SEG, Mouse_INT);
+#ifndef USE_NEW_INT
   SETIVEC(0x74, Mouse_SEG, Mouse_INT74);
+#else /* USE_NEW_INT */
+#if 0
+  SETIVEC(0x74, Mouse_SEG, Mouse_ROUTINE_OFF);
+#endif
+#endif /* USE_NEW_INT */
 
   /* Return 0xffff on mouse installed, 0x0000 - no mouse driver installed */
   /* Return number of mouse buttons */
@@ -798,8 +896,14 @@ mouse_reset(int flag)
   mouse.lrcount = mouse.mrcount = mouse.rrcount = 0;
 
   if (!mouse.ignorexy) {
-    mouse.speed_x = 8;
-    mouse.speed_y = 16;
+    if (mice->type != MOUSE_X) {
+	    mouse.speed_x = 8;
+	    mouse.speed_y = 16;
+    } else {
+	    /* default to 1 to 1 scaling for apps tha read mickeys */
+	    mouse.speed_x = 8;
+	    mouse.speed_y = 8;
+    }
   }
 
   mouse.exc_lx = mouse.exc_ux = 0;
@@ -817,8 +921,8 @@ mouse_reset(int flag)
 
   mouse.textscreenmask = 0xffff;
   mouse.textcursormask = 0x7f00;
-  memcpy(mouse.graphscreenmask,default_graphscreenmask,32);
-  memcpy(mouse.graphcursormask,default_graphcursormask,32);
+  memcpy((void *)mouse.graphscreenmask,default_graphscreenmask,32);
+  memcpy((void *)mouse.graphcursormask,default_graphcursormask,32);
   mouse.hotx = mouse.hoty = -1;
 }
 
@@ -969,6 +1073,7 @@ mouse_setxminmax(void)
   mouse.maxx = (LWORD(ecx) > LWORD(edx)) ? LWORD(ecx) : LWORD(edx);
   mouse.minx = mouse_roundx(mouse.minx);
   mouse.maxx = mouse_roundx(mouse.maxx);
+  mouse.maxx += (1 << mouse.xshift) -1;
 }
 
 void 
@@ -981,6 +1086,7 @@ mouse_setyminmax(void)
   mouse.maxy = (LWORD(ecx) > LWORD(edx)) ? LWORD(ecx) : LWORD(edx);
   mouse.miny = mouse_roundy(mouse.miny);
   mouse.maxy = mouse_roundy(mouse.maxy);
+  mouse.maxy += (1 << mouse.yshift) -1;
 }
 
 void 
@@ -999,11 +1105,11 @@ mouse_set_gcur(void)
   mouse.hoty = LWORD(ecx);
 
   /* remember mask definition */
-  memcpy(mouse.graphscreenmask,ptr,32);
-  memcpy(mouse.graphcursormask,ptr+16,32);
+  memcpy((void *)mouse.graphscreenmask,ptr,32);
+  memcpy((void *)mouse.graphcursormask,ptr+16,32);
 
   /* compile it so that it can acutally be drawn. */
-  define_graphics_cursor(mouse.graphscreenmask,mouse.graphcursormask);
+  define_graphics_cursor((short *)mouse.graphscreenmask,(short *)mouse.graphcursormask);
 }
 
 void 
@@ -1089,37 +1195,44 @@ void
 mouse_enable_internaldriver()
 {
   mice->intdrv = TRUE;
-  
   SETIVEC(0x33, Mouse_SEG, Mouse_INT);
+#ifndef USE_NEW_INT
   SETIVEC(0x74, Mouse_SEG, Mouse_INT74);
-
+#else /* USE_NEW_INT */
+#if 0
+  SETIVEC(0x74, Mouse_SEG, Mouse_ROUTINE_OFF);
+#endif
+#endif /* USE_NEW_INT */
   m_printf("MOUSE: Enable InternalDriver\n");
 }
 
-void 
-mouse_keyboard(int sc)
+
+
+/* XXX - something's wrong here. Shouldn't the first 4 cases move the cursor?
+ */
+void mouse_keyboard(int sc)
 {
   switch (sc) {
-  case 0x50:
+  case KEY_DOWN:
     mouse.rbutton = mouse.lbutton = mouse.mbutton = 0;
     mouse_move();
     break;
-  case 0x4b:
+  case KEY_LEFT:
     mouse.rbutton = mouse.lbutton = mouse.mbutton = 0;
     mouse_move();
     break;
-  case 0x4d:
+  case KEY_RIGHT:
     mouse.rbutton = mouse.lbutton = mouse.mbutton = 0;
     mouse_move();
     break;
-  case 0x48:
+  case KEY_UP:
     mouse.rbutton = mouse.lbutton = mouse.mbutton = 0;
     mouse_move();
     break;
-  case 0x47:
+  case KEY_HOME:
     mouse_lb();
     break;
-  case 0x4f:
+  case KEY_END:
     mouse_rb();
     break;
   default:
@@ -1147,6 +1260,10 @@ mouse_round_coords()
 void 
 mouse_move(void)
 {
+#if !defined(USE_NEW_INT) && (INT10_WATCHER_SEG == BIOSSEG)
+   /* With the mouse watcher can do this only on mode resets */ 
+   mouse_reset_to_current_video_mode();
+#endif
   if (mouse.x <= mouse.minx) mouse.x = mouse.minx;
   if (mouse.y <= mouse.miny) mouse.y = mouse.miny;
   if (mouse.x >= mouse.maxx) mouse.x = mouse.maxx;
@@ -1164,7 +1281,7 @@ mouse_move(void)
 	mouse_cursor(-1);
   }
 
-  m_printf("MOUSE: move. x=%x,y=%x\n", mouse.x, mouse.y);
+  m_printf("MOUSE: move: x=%d,y=%d\n", mouse.x, mouse.y);
    
   mouse_delta(DELTA_CURSOR);
 }
@@ -1223,6 +1340,83 @@ mouse_rb(void)
   }
 }
 
+void mouse_move_buttons(int lbutton, int mbutton, int rbutton)
+{
+	/* Provide 3 button emulation on 2 button mice,
+	   But only when PC Mouse Mode is set, otherwise
+	   Microsoft Mode = 2 buttons.
+	   Middle press if left/right pressed together. 
+	   Middle release done for us by the above routines. 
+	   This bit also resets the left/right button, because
+	   we can't have all three buttons pressed at once. 
+	   Well, we could, but not under emulation on 2 button mice. 
+	   Alan Hourihane */
+
+	if (mice->emulate3buttons && mouse.threebuttons) {
+		if (lbutton && rbutton) {
+			mbutton = 1;  /* Set middle button */
+		}
+	}
+	mouse.oldlbutton = mouse.lbutton;
+	mouse.oldmbutton = mouse.mbutton;
+	mouse.oldrbutton = mouse.rbutton;
+	mouse.lbutton = !!lbutton;
+	mouse.mbutton = !!mbutton;
+	mouse.rbutton = !!rbutton;
+	/*
+	 * update the event mask
+	 */
+	if (mouse.oldlbutton != mouse.lbutton)
+	   mouse_lb();
+        if (mouse.threebuttons && mouse.oldmbutton != mouse.mbutton)
+	   mouse_mb();
+	if (mouse.oldrbutton != mouse.rbutton)
+	   mouse_rb();
+}
+
+void mouse_move_relative(int dx, int dy)
+{
+	/* according to the interrupt list, the speed setting is in
+		mickeys per eight pixels. */
+	/* IDEA: running dx and dy through a filter which dampens
+		values near zero and amplifies larger values might
+		give us a cheap acceleration profile. */
+	mouse.x += ((dx << 3) / mouse.speed_x);
+	mouse.y += ((dy << 3) / mouse.speed_y);
+	mouse.mickeyx += dx;
+	mouse.mickeyy += dy;
+	/*
+	 * update the event mask
+	 */
+	if (dx || dy)
+	   mouse_move();
+}
+
+void mouse_move_absolute(int x, int y, int x_range, int y_range)
+{
+	int dx, dy, new_x, new_y, mx_range, my_range;
+	mx_range = mouse.maxx - mouse.minx +1;
+	my_range = mouse.maxy - mouse.miny +1;
+	new_x = (x*mx_range)/x_range + mouse.minx;
+	new_y = (y*my_range)/y_range + mouse.miny;
+	dx = (((new_x - mouse.x) * mouse.speed_x) >> 3);
+	dy = (((new_y - mouse.y) * mouse.speed_y) >> 3);
+	mouse.mickeyx += dx;
+	mouse.mickeyy += dy;
+	mouse.x = new_x;
+	mouse.y = new_y;
+
+	m_printf("mouse_move_absolute(%d, %d, %d, %d) -> %d %d \n",
+		 x, y, x_range, y_range, mouse.x, mouse.y);
+	m_printf("mouse_move_absolute dx:%d dy%d mickeyx%d mickeyy%d\n", 
+		 dx, dy, mouse.mickeyx, mouse.mickeyy);
+	/*
+	 * update the event mask
+	 */
+	if (dx || dy || mouse.x != new_x || mouse.y != new_y) 
+	   mouse_move();
+}
+
 void 
 fake_int(void)
 {
@@ -1301,7 +1495,13 @@ mouse_event()
     if(in_dpmi && !in_dpmi_dos_int 
           && !((mouse.cs == DPMI_SEG) && ((void *)(Bit32u)mouse.ip == DPMI_mouse_callback)))
                   fake_pm_int();
+
+    /* push iret frame on _SS:_SP. At F000:20F7 (bios.S) we get an
+     * iret and return to _CS:_IP */
     fake_int();
+
+    /* push all 16-bit regs plus _DS,_ES. At F000:20F4 (bios.S) we find
+     * 'pop es; pop ds; popa' */
     fake_pusha();
 
     LWORD(eax) = mouse_events;
@@ -1313,8 +1513,10 @@ mouse_event()
     if (mouse.threebuttons)
       LWORD(ebx) |= (mouse.mbutton ? 4 : 0);
 
+    /* push return address F000:20F4 */
     fake_call(Mouse_SEG, Mouse_OFF + 4);	/* skip to popa */
 
+    /* jump to mouse cs:ip */
     REG(cs) = mouse.cs;
     REG(eip) = mouse.ip;
 
@@ -1328,7 +1530,7 @@ mouse_event()
     m_printf("MOUSE: event %d, x %d ,y %d, mx %d, my %d, b %x\n",
 	     mouse_events, mouse.x, mouse.y, mouse.maxx, mouse.maxy, LWORD(ebx));
     m_printf("MOUSE: "
-	     "should call %04x:%04x (actually %04x:%04x)\n"
+	     "should call %04x:%04x (actually %04x:%04x)"
 	     ".........jumping to %04x:%04x\n",
 	     mouse.cs, mouse.ip, *mouse.csp, *mouse.ipp,
 	     LWORD(cs), LWORD(eip));	
@@ -1358,10 +1560,8 @@ static void
 mouse_update_cursor(void)
 {
 	/* sigh, too many programs seem to expect the mouse cursor
-		to magically redraw itself when in text mode, so we'll
-		bend to their will... */
-	if ((mouse.rx != mouse.oldrx || mouse.ry != mouse.oldry) &&
-			!mouse.gfx_cursor) {
+		to magically redraw itself, so we'll bend to their will... */
+	if (mouse.rx != mouse.oldrx || mouse.ry != mouse.oldry) {
 		mouse_do_cur();
 		mouse.oldrx = mouse.rx;
 		mouse.oldry = mouse.ry;
@@ -1467,6 +1667,7 @@ mouse_sethandler(void *f, us * cs, us * ip)
 void
 mouse_init(void)
 {
+  PRIV_SAVE_AREA
   serial_t *sptr=NULL;
   char mouse_ver[]={2,3,4,5,0x14,0x7,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,0x3f};
 #if 1 /* BUG CATCHER */
@@ -1487,22 +1688,22 @@ mouse_init(void)
   else
   	mouse.threebuttons = FALSE;
 
-  mouse_reset(1);		/* Let's set defaults now ! */
-
 #ifdef X_SUPPORT
   if (config.X) {
     mice->intdrv = TRUE;
     mice->type = MOUSE_X;
     memcpy(p,mouse_ver,sizeof(mouse_ver));
     m_printf("MOUSE: X Mouse being set\n");
-    return;
   }
+  else 
 #endif
   
   if ( ! config.usesX ){
     if (mice->intdrv) {
       m_printf("Opening internal mouse: %s\n", mice->dev);
+      enter_priv_on();  /* The mouse might need special permisions to open (esp r/w). */
       mice->fd = DOS_SYSCALL(open(mice->dev, O_RDWR | O_NONBLOCK));
+      leave_priv_setting();
       if (mice->fd == -1) {
  	mice->intdrv = FALSE;
  	mice->type = MOUSE_NONE;
@@ -1513,11 +1714,11 @@ mouse_init(void)
       add_to_io_select(mice->fd, mice->add_to_io_select);
       DOSEMUSetupMouse();
       memcpy(p,mouse_ver,sizeof(mouse_ver));
-      return;
     }
-
-    if ((mice->type == MOUSE_PS2) || (mice->type == MOUSE_BUSMOUSE)) {
+    else if ((mice->type == MOUSE_PS2) || (mice->type == MOUSE_BUSMOUSE)) {
+      enter_priv_on();
       mice->fd = open(mice->dev, O_RDWR | O_NONBLOCK);
+      leave_priv_setting();
       mice->add_to_io_select = 0;
       if (mice->fd == -1) {
 	  error("Cannot open internal mouse device %s\n",mice->dev);
@@ -1554,9 +1755,10 @@ mouse_init(void)
     add_to_io_select(mice->fd, mice->add_to_io_select);
     DOSEMUSetupMouse();
     memcpy(p,mouse_ver,sizeof(mouse_ver));
-    return;
   }
-
+  
+  /* We set the defaults at the end so that we can test the mouse type */
+  mouse_reset(1);		/* Let's set defaults now ! */
 }
 
 void
@@ -1576,13 +1778,14 @@ mouse_close(void)
     else
        m_printf("mouse_close: tcsetattr failed: %s\n",strerror(errno));
     m_printf("mouse_close: closing mouse device, fd=%d\n",mice->fd);
-    DOS_SYSCALL(close(mice->fd));
+    close(mice->fd);
     m_printf("mouse_close: ok\n");
     return;
   }
 
-  if ((mice->type == MOUSE_PS2) || (mice->type == MOUSE_BUSMOUSE))
-    DOS_SYSCALL(close(mice->fd));
+  if (((mice->type == MOUSE_PS2) || (mice->type == MOUSE_BUSMOUSE))
+	&& (mice->fd != -1))
+    close(mice->fd);
 }
 
 /* TO DO LIST: (in no particular order)
