@@ -1,6 +1,13 @@
+/* 
+ * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /*
  * DANG_BEGIN_MODULE
  *
+ * REMARK
  * Here's all the calls to the code to try and properly save & restore
  * the video state between VC's and the attempts to control updates to
  * the VC whilst the user is using another. We map between the real
@@ -11,6 +18,7 @@
  * the emulator try to execute the card's initialization routine which is
  * normally located at address c000:0003. This can now be changed as an
  * option.
+ * /REMARK
  *
  * DANG_END_MODULE
  *
@@ -19,7 +27,7 @@
  *
  */
 
-
+#include <features.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/ioctl.h>
@@ -34,23 +42,33 @@
 #include <sys/mman.h>
 #endif
 #ifdef __linux__
+#ifndef __GLIBC__
 #define MAP_ANON MAP_ANONYMOUS
 extern caddr_t mmap __P ((caddr_t __addr, size_t __len,
 			  int __prot, int __flags, int __fd, off_t __off));
 extern int munmap __P ((caddr_t __addr, size_t __len));
 #include <linux/mman.h>
+#else
+#include <sys/mman.h>
+#endif
 #endif
 #include <signal.h>
 #include <sys/stat.h>
 #ifdef __linux__
+#if __GLIBC__ > 1
+#include <sys/vt.h>
+#include <sys/kd.h>
+#else
 #include <linux/vt.h>
 #include <linux/kd.h>
 #endif
+#endif
 #ifdef __NetBSD__
 #include <machine/pcvt_ioctl.h>
 #endif
 
 #define INIT_C2TRAP
+#define NO_VC_NO_DEBUG
 
 #include "emu.h"
 #include "memory.h"
@@ -76,8 +94,6 @@ extern inline void console_update_cursor (int, int, int, int);
 static void set_dos_video ();
 void put_video_ram ();
 
-static struct debug_flags save_debug_flags;
-
 extern int
   dosemu_sigaction (int sig, struct sigaction *, struct sigaction *);
 static void SIGRELEASE_call (void);
@@ -86,10 +102,10 @@ static void SIGACQUIRE_call (void);
 void dump_video_regs (void);
 
 #ifdef WANT_DUMP_VIDEO
-static void dump_video ();
+static void dump_video (void);
 #endif
 
-void dump_video_linux ();
+void dump_video_linux (void);
 
 
 static int  color_text;
@@ -119,7 +135,7 @@ forbid_switch (void)
 /* check whether we are running on the console; initialise
  * config.console, console_fd and scr_state.console_no
  */
-void check_console() {
+void check_console(void) {
 
 #ifdef __NetBSD__
     struct screeninfo info;
@@ -147,7 +163,7 @@ void check_console() {
     minor = chkbuf.st_rdev & 0xff;
 
     /* console major num is 4, minor 64 is the first serial line */
-    if ((major == 4) && (minor < 64)) {
+    if (can_do_root_stuff && (major == 4) && (minor < 64)) {
        scr_state.console_no = minor;
        config.console=1;
     }
@@ -192,6 +208,7 @@ parent_close_mouse (void)
 static void
 parent_open_mouse (void)
 {
+  PRIV_SAVE_AREA
   if (mice->intdrv)
     {
       enter_priv_on(); /* The mouse may not be a resource everyone can open. */
@@ -223,7 +240,11 @@ acquire_vt (int sig, struct sigcontext_struct context)
 {
   dos_has_vt = 1;
 
-  d = save_debug_flags;
+#ifdef NO_VC_NO_DEBUG
+  if (user_vc_switch) shut_debug = 0;
+#endif
+  flush_log();
+
   v_printf ("VID: Acquiring VC\n");
   forbid_switch ();
 #ifdef __NetBSD__
@@ -237,26 +258,12 @@ acquire_vt (int sig, struct sigcontext_struct context)
   scr_state.current = 1;
 }
 
-static inline void
-get_permissions (void)
-{
-  DOS_SYSCALL (set_ioperm (0x3b0, 0x3df - 0x3b0 + 1, 1));
-}
-
-static inline void
-giveup_permissions (void)
-{
-  DOS_SYSCALL (set_ioperm (0x3b0, 0x3df - 0x3b0 + 1, 0));
-}
-
 static void
 set_dos_video (void)
 {
   if (!config.vga)
     return;
 
-  v_printf ("Setting DOS video: gfx_mode: %d\n", gfx_mode);
-
   /* jesx */
   /* After all that fun up there, get permissions and save/restore states */
   if (video_initialized)
@@ -302,8 +309,6 @@ set_linux_video (void)
 static void
 SIGRELEASE_call (void)
 {
-  save_debug_flags = d;
-
   if (scr_state.current == 1)
     {
       v_printf ("VID: Releasing VC\n");
@@ -354,7 +359,9 @@ SIGRELEASE_call (void)
   else {
    /* we don't want any debug activity while VC is not active */
     flush_log();
-    memset(&d, 0, sizeof(struct debug_flags));
+#ifdef NO_VC_NO_DEBUG
+    if (user_vc_switch) shut_debug = 1;
+#endif
   }
 }
 
@@ -365,7 +372,7 @@ wait_vc_active (void)
     {
       if (errno != EINTR)
 		{
-			error ("ERROR: VT_WAITACTIVE for %d gave %d: %s\n", scr_state.console_no,
+			error("VT_WAITACTIVE for %d gave %d: %s\n", scr_state.console_no,
 			     errno, strerror (errno));
 	    }
       return -1;
@@ -481,7 +488,6 @@ get_video_ram (int waitflag)
 
       close_kmem ();
       /* the code below is done by the video save/restore code */
-      get_permissions ();
     }
   else
     {
@@ -514,7 +520,7 @@ get_video_ram (int waitflag)
 
       if ((long) graph_mem < 0)
 	{
-	  error ("ERROR: mmap error in get_video_ram (text): %lx, errno %d\n",
+	  error("mmap error in get_video_ram (text): %x, errno %d\n",
 		 (Bit32u)graph_mem, errno);
 	  return;
 	}
@@ -522,7 +528,6 @@ get_video_ram (int waitflag)
 	v_printf ("CONSOLE VIDEO address: %p %p %p\n", (void *) graph_mem,
 		  (void *) phys_text_base, (void *) PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE)));
 
-      get_permissions ();
       /* copy contents of page onto video RAM */
       memcpy ((caddr_t) PAGE_ADDR (READ_BYTE(BIOS_CURRENT_SCREEN_PAGE)), textbuf, TEXT_SIZE);
     }
@@ -575,7 +580,6 @@ put_video_ram (void)
 	  memcpy (scr_state.virt_address, putbuf, TEXT_SIZE);
 	}
 
-      giveup_permissions ();
       scr_state.mapped = 0;
     }
   else
@@ -614,6 +618,7 @@ set_process_control (void)
 void
 clear_process_control (void)
 {
+  PRIV_SAVE_AREA
   struct vt_mode vt_mode;
 
   vt_mode.mode = VT_AUTO;
@@ -632,6 +637,7 @@ static u_char kmem_open_count = 0;
 void
 open_kmem (void)
 {
+  PRIV_SAVE_AREA
   /* as I understad it, /dev/kmem is the kernel's view of memory,
      * and /dev/mem is the identity-mapped (i.e. physical addressed)
      * memory. Currently under Linux, both are the same.
@@ -651,7 +657,7 @@ open_kmem (void)
   leave_priv_setting();
   if (mem_fd < 0)
     {
-      error ("ERROR: can't open /dev/mem: errno=%d, %s \n",
+      error("can't open /dev/mem: errno=%d, %s \n",
 	     errno, strerror (errno));
       leavedos (0);
       return;
@@ -729,15 +735,29 @@ get_perm (void)
 	  v_printf ("VGA: can't get I/O permissions \n");
 	  exit (-1);
 	}
-      if (((config.chipset==S3) || (config.chipset==CIRRUS)) &&
-		(set_ioperm (0x102, 1, 1) || set_ioperm (0x2ea, 4, 1)))
-	{
-	  v_printf ("S3/CIRRUS: can't get I/O permissions \n");
+      if ((config.chipset == S3) &&
+	(set_ioperm(0x102, 1, 1) || set_ioperm(0x2ea, 4, 1))) {
+	  v_printf("S3: can't get I/O permissions");
 	  exit (-1);
 	}
-      if (config.chipset == ATI && (set_ioperm (0x102, 1, 1) || set_ioperm (0x1ce, 2, 1) || set_ioperm (0x2ec, 4, 1)))
-	{
-	  v_printf ("ATI: can't get I/O permissions \n");
+      if ((config.chipset == CIRRUS) &&
+	(set_ioperm(0x102, 2, 1) || set_ioperm(0x2ea, 4, 1))) {
+	  v_printf("CIRRUS: can't get I/O permissions");
+	  exit (-1);
+	}
+      if ((config.chipset == ATI) &&
+	(set_ioperm(0x102, 1, 1) || set_ioperm(0x1ce, 2, 1) || set_ioperm(0x2ec, 4, 1))) {
+	  v_printf("ATI: can't get I/O permissions");
+	  exit (-1);
+	}
+      if ((config.chipset == MATROX) &&
+	(set_ioperm(0x102, 2, 1) || set_ioperm(0x2ea, 4, 1) || set_ioperm(0x3de, 2, 1))) {
+	  v_printf("MATROX: can't get I/O permissions");
+	  exit (-1);
+	}
+      if ((config.chipset == WDVGA) &&
+	(set_ioperm(0x102, 2, 1) || set_ioperm(0x3de, 2, 1))) {
+	  v_printf("WDVGA: can't get I/O permissions");
 	  exit (-1);
 	}
       /* color or monochrome text emulation? */
@@ -795,16 +815,29 @@ release_perm (void)
 	      v_printf ("VGA: can't release I/O permissions \n");
 	      leavedos (-1);
 	    }
-	  if (((config.chipset==S3) || (config.chipset==CIRRUS)) && 
-		(set_ioperm (0x102, 1, 0) || set_ioperm (0x2ea, 4, 0)))
-	    {
-	      v_printf ("S3/CIRRUS: can't release I/O permissions\n");
+	  if ((config.chipset == S3) &&
+		(set_ioperm(0x102, 1, 0) || set_ioperm(0x2ea, 4, 0))) {
+	      v_printf ("S3: can't release I/O permissions\n");
 	      leavedos (-1);
 	    }
-	  if (config.chipset == ATI && 
-		(set_ioperm (0x102, 1, 0) || set_ioperm (0x1ce, 2, 0) || set_ioperm (0x2ec, 4, 0)))
-	    {
-	      v_printf ("ATI: can't release I/O permissions \n");
+	  if ((config.chipset == CIRRUS) &&
+		(set_ioperm(0x102, 2, 0) || set_ioperm(0x2ea, 4, 0))) {
+	      v_printf ("CIRRUS: can't release I/O permissions\n");
+	      leavedos (-1);
+	    }
+	  if ((config.chipset == ATI) &&
+		(set_ioperm(0x102, 1, 0) || set_ioperm(0x1ce, 2, 0) || set_ioperm(0x2ec, 4, 0))) {
+	      v_printf ("ATI: can't release I/O permissions\n");
+	      leavedos (-1);
+	    }
+	  if ((config.chipset == MATROX) &&
+		(set_ioperm(0x102, 2, 0) || set_ioperm(0x2ea, 4, 0) || set_ioperm(0x3de, 2, 0))) {
+	      v_printf ("MATROX: can't release I/O permissions\n");
+	      leavedos (-1);
+	    }
+	  if ((config.chipset == WDVGA) &&
+		(set_ioperm(0x102, 2, 0) || set_ioperm(0x3de, 2, 0))) {
+	      v_printf ("WDVGA: can't release I/O permissions\n");
 	      leavedos (-1);
 	    }
 	}
@@ -899,10 +932,10 @@ reset_att (void)
 /* Attempt to virtualize calls to video ports */
 
 u_char att_d_index = 0;
-static isr_read = 0;
+static int isr_read = 0;
 
 u_char
-video_port_in (int port)
+video_port_in (ioport_t port)
 {
   /* v_printf("Video read on port 0x%04x.\n",port); */
   switch (port)
@@ -986,7 +1019,7 @@ video_port_in (int port)
 }
 
 void
-video_port_out (u_char value, int port)
+video_port_out (ioport_t port, u_char value)
 {
   /* v_printf("Video write on port 0x%04x,byte 0x%04x.\n",port, value); */
   switch (port)
@@ -1004,7 +1037,7 @@ video_port_out (u_char value, int port)
 	  dosemu_regs.regs[CRT + dosemu_regs.regs[CRTI]] = value;
 	}
       else
-	ext_video_port_out (value, port);
+	ext_video_port_out (port, value);
       break;
     case GRA_I:
       v_printf ("Set Index GRAI 0x%02x\n", value);
@@ -1017,7 +1050,7 @@ video_port_out (u_char value, int port)
 	  dosemu_regs.regs[GRA + dosemu_regs.regs[GRAI]] = value;
 	}
       else
-	ext_video_port_out (value, port);
+	ext_video_port_out (port, value);
       break;
     case SEQ_I:
       v_printf ("Set Index SEQI 0x%02x\n", value);
@@ -1030,7 +1063,7 @@ video_port_out (u_char value, int port)
 	  dosemu_regs.regs[SEQ + dosemu_regs.regs[SEQI]] = value;
 	}
       else
-	ext_video_port_out (value, port);
+	ext_video_port_out (port, value);
       break;
     case ATT_IW:
       if (isr_read)
@@ -1043,7 +1076,7 @@ video_port_out (u_char value, int port)
 	{
 	  isr_read = 1;
 	  if (att_d_index > 20)
-	    ext_video_port_out (value, port);
+	    ext_video_port_out (port, value);
 	  else
 	    {
 	      v_printf ("Write Data at ATT Index 0x%02x = 0x%02x \n", att_d_index, value);
@@ -1081,7 +1114,7 @@ video_port_out (u_char value, int port)
       dosemu_regs.regs[GR2P] = value;
       break;
     default:
-      ext_video_port_out (value, port);
+      ext_video_port_out (port, value);
     }
   return;
 }
@@ -1173,12 +1206,9 @@ dump_video_linux (void)
 /*
  * install_int_10_handler - install a handler for the video-interrupt (int 10)
  *                          at address INT10_SEG:INT10_OFFS. Currently
- *                          it's f100:4100.
+ *                          it's f800:4200.
  *                          The new handler is only installed, if the bios
- *                          handler at f100:4100 is not the appropriate on
- *                          that means, if we use not mda with X
- *                          The new handler is only installed, if the bios
- *                          handler at f100:4100 is not the appropriate on
+ *                          handler at f800:4200 is not the appropriate on
  *                          that means, if we use not mda with X
  */
 
@@ -1200,7 +1230,7 @@ install_int_10_handler (void)
       *ptr++ = 0x08;		/* mov  al,08  ; Start Video init     */
 
       *ptr++ = 0xcd;
-      *ptr++ = 0xe6;		/* int  e6          */
+      *ptr++ = DOS_HELPER_INT;		/* int  e6          */
 
       if (config.vbios_seg == 0xe000)
 	{
@@ -1222,7 +1252,7 @@ install_int_10_handler (void)
       *ptr++ = 0x09;		/* move al,09   ; Finished video init    */
 
       *ptr++ = 0xcd;
-      *ptr++ = 0xe6;		/* int  e6          */
+      *ptr++ = DOS_HELPER_INT;		/* int  e6          */
 
       *ptr++ = 0x58;		/* pop ax           */
       *ptr++ = 0xfb;		/* start interrupts STI  */
@@ -1233,6 +1263,9 @@ install_int_10_handler (void)
     extern void bios_f000_int10ptr(), bios_f000();
     ptr = (u_char *)((BIOSSEG << 4) + ((long)bios_f000_int10ptr - (long)bios_f000));
     *((long *)ptr) = 0xe0000003;
+    v_printf("VID: new int10 handler at %p\n",ptr);
   }
+  else
+    v_printf("VID: install_int_10_handler: do nothing\n");
 #endif
 }
