@@ -1,3 +1,9 @@
+/* 
+ * (C) Copyright 1992, ..., 1998 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /*
  * video/vga.c - This file contains function for VGA-cards only 
  */
@@ -16,7 +22,12 @@
 
 #include "et4000.h"
 #include "s3.h"
+#include "ati.h"
 #include "trident.h"
+#include "avance.h"
+#include "cirrus.h"
+#include "matrox.h"
+#include "wdvga.h"
 #if 0
 #include "hgc.h"
 #endif
@@ -58,13 +69,13 @@ void set_bank_write_dummy(u_char bank)
   return;
 }
 
-u_char dummy_ext_video_port_in(int port)
+u_char dummy_ext_video_port_in(ioport_t port)
 {
   v_printf("Bad Read on port 0x%04x\n", port);
   return (0);
 }
 
-void dummy_ext_video_port_out(u_char value, int port)
+void dummy_ext_video_port_out(ioport_t port, u_char value)
 {
   v_printf("Bad Write on port 0x%04x with value 0x%02x\n", port, value);
 }
@@ -251,7 +262,7 @@ void store_vga_mem(u_char * mem, u_char mem_size[], u_char banks)
 
 /*      memcpy((caddr_t) (mem + position), (caddr_t) GRAPH_BASE, p[plane]); */
       MEMCPY_2UNIX((caddr_t) (mem + position), (caddr_t) GRAPH_BASE, p[plane]);
-      v_printf("READ Bank=%d, plane=0x%02x, *mem=0x%x, GRAPH_BASE=%08x, mem=0x%lx\n",
+      v_printf("READ Bank=%d, plane=0x%02x, *mem=0x%x, GRAPH_BASE=%08x, mem=0x%x\n",
 	 banks, plane, *(int *)(mem + position), *(int *) GRAPH_BASE, (Bit32u) (mem + position));
       position = position + p[plane];
     }
@@ -463,6 +474,14 @@ int vga_initialize(void)
   dosemu_regs.mem = NULL;
   get_perm();
 
+  /* defaults to override */
+  save_ext_regs = save_ext_regs_dummy;
+  restore_ext_regs = restore_ext_regs_dummy;
+  set_bank_read = set_bank_read_dummy;
+  set_bank_write = set_bank_write_dummy;
+  ext_video_port_in = dummy_ext_video_port_in;
+  ext_video_port_out = dummy_ext_video_port_out;
+
   switch (config.chipset) {
   case TRIDENT:
     vga_init_trident();
@@ -476,13 +495,28 @@ int vga_initialize(void)
     vga_init_s3();
     v_printf("S3 CARD in use\n");
     break;
+  case AVANCE:
+    vga_init_avance();
+    v_printf("Avance Logic CARD in use\n");
+    break;
+  case ATI:
+    vga_init_ati();
+    v_printf("ATI CARD in use\n");
+    break;
+  case CIRRUS:
+    vga_init_cirrus();
+    v_printf("Cirrus CARD in use\n");
+    break;
+  case MATROX:
+    vga_init_matrox();
+    v_printf("Matrox CARD in use\n");
+    break;
+  case WDVGA:
+    vga_init_wd();
+    v_printf("Paradise CARD in use\n");
+    break;
+
   default:
-    save_ext_regs = save_ext_regs_dummy;
-    restore_ext_regs = restore_ext_regs_dummy;
-    set_bank_read = set_bank_read_dummy;
-    set_bank_write = set_bank_write_dummy;
-    ext_video_port_in = dummy_ext_video_port_in;
-    ext_video_port_out = dummy_ext_video_port_out;
     v_printf("Unspecific VIDEO selected = 0x%04x\n", config.chipset);
   }
 
@@ -632,7 +666,7 @@ void init_vga_card(void)
 #endif
 
   if (!config.mapped_bios) {
-    error("ERROR: CAN'T DO VIDEO INIT, BIOS NOT MAPPED!\n");
+    error("CAN'T DO VIDEO INIT, BIOS NOT MAPPED!\n");
     return;
   }
   if (set_ioperm(0x3b0, 0x3db - 0x3b0, 1))
